
/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <my_global.h>
#include <my_sys.h>
#include <assert.h>
#include <my_dir.h>
#include <m_string.h>

#define RAID_SIZE_UNKNOWN ~(my_off_t) 0
#define RAID_SEEK_DONE ~(off_t) 0

typedef struct my_chunk_pos
{
  ulong total_block;
  uint  this_block;
  uint  remaining_bytes;
} MY_CHUNK_POS;

inline static void raid_calculate(MY_FILE * my_file, MY_CHUNK_POS * chunk_pos)
{
  DBUG_ENTER("raid_calculate begin");
  DBUG_PRINT("info",("position: %lu my_file->raid_chunksize: %d, size: %lu",
		(ulong) my_file->pos_in_file, my_file->raid_root.raid_chunksize,
		(ulong) my_file->end_of_file));
  chunk_pos->total_block = (ulong) (my_file->pos_in_file /
			my_file->raid_root.raid_chunksize);
  chunk_pos->this_block = chunk_pos->total_block % 
			my_file->raid_root.raid_chunks;
  chunk_pos->remaining_bytes = (uint) (my_file->raid_root.raid_chunksize -
			(my_file->pos_in_file - chunk_pos->total_block * 
			my_file->raid_root.raid_chunksize));
  DBUG_PRINT("info", ("total_block: %d  this_block: %d  remaining_bytes:%d",
			chunk_pos->total_block, chunk_pos->this_block,
			chunk_pos->remaining_bytes));
  DBUG_VOID_RETURN;
}

MY_FILE_VMT raid_file_vmt =
{
  0,
  &my_u_raid_read,
  &my_u_raid_pread,
  &my_u_raid_write,
  &my_u_raid_pwrite,
  &my_u_raid_close,
  &my_u_raid_lock,
  &my_u_raid_chsize,
  &my_u_raid_fstat,
  &my_u_raid_tell,
  &my_u_raid_seek,
  &my_u_raid_free
};

void my_u_raid_free(MY_FILE *my_file)
{
}
                        
bool init_my_file_for_raid(MY_FILE * my_file, uint raid_type, uint raid_chunks, 
			   ulong raid_chunksize)
{
  MY_FILE_RAID_ROOT *my_raid_root;
  MY_CHUNK *my_chunks;
  
  DBUG_ENTER("init_my_file_for_raid");
  my_chunks= (MY_CHUNK*)my_malloc(sizeof(MY_CHUNK)*raid_chunks, MYF(0));
  if (!my_chunks)
    DBUG_RETURN(1);				 /* Not enough memory */

  my_raid_root= &my_file->raid_root;
  my_raid_root->raid_vector= my_chunks;
  my_file->pos_in_file= 0;
  my_raid_root->raid_type= raid_type;
  my_raid_root->raid_chunksize= raid_chunksize;
  my_raid_root->raid_chunks= raid_chunks;
  my_file->vmt= &raid_file_vmt;
  DBUG_RETURN(0);
}


bool intern_my_raid_create(MY_FILE * my_file, int CreateFlags, int access_flags,
			   myf MyFlags)
{
  MY_CHUNK *my_chunks;
  char file_name[FN_REFLEN];
  char dir_name[FN_REFLEN];
  uint i;
  MY_STAT status;
  int save_errno;

  DBUG_ENTER("intern_my_raid_create");
  DBUG_PRINT("enter",
	     ("FileName: %s  CreateFlags: %d  access_flags: %d  MyFlags: %d",
	     my_file->name.buffer, CreateFlags, access_flags, MyFlags));

  strmake(dir_name, my_file->name.buffer, MF_DIR_LEN(my_file));
  my_chunks= my_file->raid_root.raid_vector;
  i= my_file->raid_root.raid_chunks - 1;
  do
  {
    /* Create subdir */
    (void)sprintf(file_name, "%s%02x", dir_name, i);
    unpack_dirname(file_name, file_name);
    if (my_stat(file_name, &status, MYF(0)))
    {
      DBUG_PRINT("info",("%02x exists, skipping directory creation",i));
    }
    else
    {
      if (my_mkdir(file_name, 0777, MYF(0)))
      {
        DBUG_PRINT("error",("mkdir failed for %d",i));
        goto error;
      }
    }
    
    /* Create file */
    sprintf(file_name,"%s%02x/%s", dir_name, i, my_file->name.base);
    if (!((my_chunks+i)->my_file=(MY_FILE*) init_my_file((my_chunks+i)->my_file,
	file_name, MY_USE_ALLOCATED_BUFFER|USE_ALLOCATED_MY_FILE)) ||
        my_simple_create((my_chunks + i)->my_file, CreateFlags,
	access_flags, (myf)MyFlags))
      goto error;
    (my_chunks+i)->pos_in_chunk=RAID_SEEK_DONE;      
  } while (i--);
  
  my_file->end_of_file= 0;
  DBUG_RETURN(0);			/* Raid file created successfully */

error:
  {
    save_errno=my_errno;
    while (++i < my_file->raid_root.raid_chunks)
    {
      unpack_filename(file_name, (my_chunks+i)->my_file->name.buffer);
      my_simple_close((my_chunks+i)->my_file, MYF(0));
      my_simple_delete((my_chunks+i)->my_file, MYF(0));
      free_my_file((my_chunks+i)->my_file);
    }
    my_errno=save_errno;
  }
  DBUG_RETURN(-1);
}

bool intern_my_raid_open(MY_FILE * my_file, int Flags, myf MyFlags )
{ 
  MY_CHUNK *my_chunks;
  uint i;
  char dir_name[FN_REFLEN];
  char file_name[FN_REFLEN];
  int save_errno;
  
  DBUG_ENTER("intern_my_raid_open");
  DBUG_PRINT("enter",("FileName: %s  Flags: %d  MyFlags: %d",
		   my_file->name.buffer, Flags, MyFlags));

  my_chunks= my_file->raid_root.raid_vector;
  strmake(dir_name, my_file->name.buffer, MF_DIR_LEN(my_file));
  for(i= 0; i < my_file->raid_root.raid_chunks; i++)
  {
    (void)sprintf(file_name,"%s%02x/%s", dir_name, i, my_file->name.base);
    if (!((my_chunks+i)->my_file= (MY_FILE*)init_my_file((my_chunks+i)->my_file,
	file_name, MY_USE_ALLOCATED_BUFFER|USE_ALLOCATED_MY_FILE)) ||
        my_simple_open((my_chunks+i)->my_file, Flags, (myf)MyFlags))
    {
      save_errno= my_errno;
      while (i-- != 0)
      {
	my_simple_close((my_chunks+i)->my_file, MYF(0));
	free_my_file((my_chunks+i)->my_file);
      }
      my_errno= save_errno;
      DBUG_RETURN(-1);
    }
    (my_chunks+i)->pos_in_chunk=RAID_SEEK_DONE;
  }

  my_file->end_of_file= RAID_SIZE_UNKNOWN;
  my_u_raid_seek(my_file, 0L, MY_SEEK_END, MYF(0));

  DBUG_PRINT("info",("MYD file logical size: %llu", my_file->end_of_file));
  DBUG_RETURN(0);			/* Raid file opened successfully */
}

void calc_pos_in_chunk(MY_FILE * my_file, my_off_t pos_in_file)
{
  MY_CHUNK_POS chunk_pos;
  
  off_t rounds;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;
  MY_CHUNK *pos= my_raid_root->raid_vector;
  MY_CHUNK *end;

  DBUG_ENTER("intern_my_raid_open");
  my_file->pos_in_file= pos_in_file;
  raid_calculate(my_file, &chunk_pos);
  rounds= chunk_pos.total_block / my_raid_root->raid_chunks;
  rounds*= my_raid_root->raid_chunksize;

  for (end= pos + chunk_pos.this_block; pos < end; pos++)
    pos->pos_in_chunk= rounds + my_raid_root->raid_chunksize;
    
  (pos++)->pos_in_chunk= (rounds + my_raid_root->raid_chunksize +
			  chunk_pos.remaining_bytes);
			  
  for (end= pos + my_raid_root->raid_chunks; pos < end; pos++)
    pos->pos_in_chunk= rounds;

  DBUG_VOID_RETURN;
}

my_off_t my_u_raid_seek(MY_FILE * my_file, my_off_t pos, int whence, myf MyFlags)
{
  MY_CHUNK *my_chunks;
  uint i;       
  off_t _rounds;
  my_off_t newpos;
  MY_CHUNK_POS chunk_pos;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;

  DBUG_ENTER("my_u_raid_seek");
  DBUG_PRINT("enter",("Pos: %lu  Whence: %d  MyFlags: %d",
                   (ulong) pos, whence, MyFlags));
  my_chunks= my_raid_root->raid_vector;

  switch (whence) {
  case MY_SEEK_CUR:
    if (pos)
    {
      calc_pos_in_chunk(my_file, my_file->pos_in_file + pos);
    }
    break;
  case MY_SEEK_SET:
    if ( my_file->pos_in_file != pos )
    {
      calc_pos_in_chunk(my_file, pos);
    }
    break;
  case MY_SEEK_END:
    if (my_file->end_of_file == RAID_SIZE_UNKNOWN)
    {                                                                      
      my_file->pos_in_file = 0;
      for (i= 0; i < my_raid_root->raid_chunks ; i++ )
      {
        newpos= my_simple_seek((my_chunks + i)->my_file, 0L, MY_SEEK_END, MyFlags);
        if (newpos == MY_FILEPOS_ERROR)
                 DBUG_RETURN ((uint)MY_FILEPOS_ERROR);
        (my_chunks + i)->pos_in_chunk= RAID_SEEK_DONE;
        my_file->pos_in_file+= newpos;
      }
      my_file->end_of_file= my_file->pos_in_file;
    }
    else if (my_file->pos_in_file != my_file->end_of_file)
    {
      calc_pos_in_chunk(my_file, my_file->end_of_file);
//      my_file->pos_in_file=  my_file->end_of_file;
    }
  }
  DBUG_PRINT("On exit",("Pos: %d", my_file->pos_in_file));
  DBUG_RETURN((uint)my_file->pos_in_file);
}

size_t my_u_raid_write(MY_FILE * my_file, const byte *Buffer,
                    uint Count, myf MyFlags)
{
  const byte *bufptr = Buffer;
  uint res, GotBytes, ReadNowCount;
  MY_CHUNK_POS chunk_pos;
  MY_CHUNK *my_chunks;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;
  
  DBUG_ENTER("my_u_raid_write");
  DBUG_PRINT("enter",("Count: %d  MyFlags: %d", Count, MyFlags));

  res= 0;
  my_chunks= my_raid_root->raid_vector;
  /* Loop until data is written */
  do {
    raid_calculate(my_file, &chunk_pos);
    /* Do seeks when neccessary */
    if ((int)(my_chunks + chunk_pos.this_block)->pos_in_chunk != 
                                  RAID_SEEK_DONE)
    {
      if (my_simple_seek((my_chunks + chunk_pos.this_block)->my_file,
          (my_chunks + chunk_pos.this_block)->pos_in_chunk,
		  MY_SEEK_SET,
		  MyFlags) == MY_FILEPOS_ERROR)
        DBUG_RETURN(-1);
      (my_chunks + chunk_pos.this_block)->pos_in_chunk= RAID_SEEK_DONE;
    }
    ReadNowCount= min(Count, chunk_pos.remaining_bytes);
    GotBytes= my_simple_write((my_chunks + chunk_pos.this_block)->my_file,
                bufptr, ReadNowCount,MyFlags);
    DBUG_PRINT("loop",("Wrote bytes: %d", GotBytes));
    if (GotBytes == MY_FILE_ERROR)
      DBUG_RETURN(-1);
    res+= GotBytes;
    if (MyFlags & (MY_NABP | MY_FNABP))
      GotBytes= ReadNowCount;
    bufptr+= GotBytes;
    Count-= GotBytes;
    my_file->pos_in_file+= GotBytes;
  } while(Count);
  set_if_bigger(my_file->end_of_file, my_file->pos_in_file);
  DBUG_RETURN(res);
}

size_t my_u_raid_read(MY_FILE * my_file, byte *Buffer, uint Count, 
             myf MyFlags)
{
  byte *bufptr= (byte *)Buffer;
  uint res, GotBytes, ReadNowCount;
  MY_CHUNK_POS chunk_pos;
  MY_CHUNK *my_chunks;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;
  
  DBUG_ENTER("my_u_raid_read");
  DBUG_PRINT("enter",("Count: %d  MyFlags: %d", Count, MyFlags));

  res= 0;
  my_chunks= my_raid_root->raid_vector;
  while (Count)
  {
    raid_calculate(my_file, &chunk_pos);
    /* Do seek when neccessary */
    if ((int)(my_chunks + chunk_pos.this_block)->pos_in_chunk != 
        RAID_SEEK_DONE)
    {
      if (my_simple_seek((my_chunks + chunk_pos.this_block)->my_file,
          (my_chunks + chunk_pos.this_block)->pos_in_chunk,
		  MY_SEEK_SET,
		  MyFlags) == MY_FILEPOS_ERROR)
        DBUG_RETURN(-1);
      (my_chunks + chunk_pos.this_block)->pos_in_chunk=RAID_SEEK_DONE;
    }
    /* and read */
    ReadNowCount= min(Count, chunk_pos.remaining_bytes);
    GotBytes= my_simple_read((my_chunks + chunk_pos.this_block)->my_file, 
               bufptr, ReadNowCount, MyFlags & ~(MY_NABP | MY_FNABP));
    DBUG_PRINT("loop", ("Got bytes: %u", GotBytes));
    if (GotBytes == MY_FILE_ERROR)
      DBUG_RETURN(-1);
    if (!GotBytes)				/* End of file. */
    {
      DBUG_RETURN((MyFlags & (MY_NABP | MY_FNABP)) ? -1 : (int) res);
    }
    res+= GotBytes;
    bufptr+= GotBytes;
    Count-=  GotBytes;
    my_file->pos_in_file+= GotBytes;
  }
  DBUG_RETURN(res);
}

my_off_t my_u_raid_tell(MY_FILE * my_file, myf MyFlags)
{
  DBUG_ENTER("my_u_raid_tell");
  DBUG_PRINT("enter", ("MyFlags: %d pos_in_file %d", MyFlags, 
                my_file->pos_in_file));
  DBUG_RETURN(my_file->pos_in_file);
}

int my_u_raid_lock(MY_FILE * my_file, int locktype, my_off_t start, 
           my_off_t length, myf MyFlags)
{
  my_off_t bufptr;
  MY_CHUNK_POS chunk_pos;
  MY_CHUNK *my_chunks;
  uint ReadNowCount, GotBytes, i;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;
  
  DBUG_ENTER("my_u_raid_lock");
  DBUG_PRINT("enter",("locktype: %d  start: %lu  length: %lu  MyFlags: %d",
		      locktype, start, length, MyFlags));
		      
  bufptr= start;
  /* Loop until all data is locked */
  while(length)
  {
    raid_calculate(my_file, &chunk_pos);
    my_chunks= my_raid_root->raid_vector;
    for (i = chunk_pos.this_block; 
         (i < my_raid_root->raid_chunks) && length; i++ )
    {
      ReadNowCount= (uint)min(length, chunk_pos.remaining_bytes);
      GotBytes= my_simple_lock((my_chunks + i)->my_file, locktype, bufptr, 
            ReadNowCount, MyFlags);
      if ((int) GotBytes == -1)
	DBUG_RETURN(-1);
      bufptr+= ReadNowCount;
      length-= ReadNowCount;
      raid_calculate(my_file, &chunk_pos);
    }
  }
  DBUG_RETURN(0);
}

int my_u_raid_chsize(MY_FILE * my_file, my_off_t newlength, int filler, myf MyFlags)
{
  int newpos;
  uint rounds, i;
  MY_CHUNK_POS chunk_pos;
  MY_CHUNK *my_chunks;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;

  DBUG_ENTER("my_u_raid_chsize");
  DBUG_PRINT("enter",("Fd: %d, newlength: %d, MyFlags: %d",
                my_file->file_descr, newlength,MyFlags));

  my_chunks= my_raid_root->raid_vector;
  my_file->pos_in_file= newlength;
  raid_calculate(my_file, &chunk_pos);
  rounds= chunk_pos.total_block / my_raid_root->raid_chunks;
  for (i= 0; i < my_raid_root->raid_chunks ; i++ )
  {
    if (i < chunk_pos.this_block)
      newpos= my_simple_chsize((my_chunks + i)->my_file,
        		(rounds + 1) * my_raid_root->raid_chunksize, filler, MyFlags);
    else if (i == chunk_pos.this_block)
      newpos= my_simple_chsize((my_chunks + i)->my_file,
			rounds * my_raid_root->raid_chunksize + 
			(newlength % my_raid_root->raid_chunksize), filler, MyFlags);
    else
      newpos= my_simple_chsize((my_chunks + i)->my_file,
			rounds * my_raid_root->raid_chunksize, filler, MyFlags);
    if (newpos)
      DBUG_RETURN(1);
  }
  DBUG_RETURN(0);
}

int my_u_raid_fstat(MY_FILE * my_file, struct my_stat * sstat_area, myf MyFlags )
{
  uint i;
//  int error;
  MY_STAT status;
  MY_STAT *stat_area= (MY_STAT*)sstat_area;
  MY_CHUNK *my_chunks;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;

  DBUG_ENTER("my_u_raid_fstat");
  DBUG_PRINT("enter",("fd: %d MyFlags: %d", my_file->file_descr, MyFlags));

//  error=0;
  stat_area->st_size=0;
  stat_area->st_mtime=0;
  stat_area->st_atime=0;
  stat_area->st_ctime=0;

  my_chunks= my_raid_root->raid_vector;
  for(i=0 ; i < my_raid_root->raid_chunks ; i++)
  {
    if (my_simple_fstat((my_chunks+i)->my_file, (struct my_stat *)&status, MyFlags))
      DBUG_RETURN(1);
    stat_area->st_size+=status.st_size;
    set_if_bigger(stat_area->st_mtime,status.st_mtime);
    set_if_bigger(stat_area->st_atime,status.st_atime);
    set_if_bigger(stat_area->st_ctime,status.st_ctime);
    DBUG_PRINT("File size",("Size: %d", stat_area->st_size));
  }
  DBUG_RETURN(0);
}

int my_u_raid_close(MY_FILE * my_file, myf MyFlags)
{
  int err, save_err;
  uint i;
  MY_CHUNK *my_chunks;
  MY_FILE_RAID_ROOT *my_raid_root = &my_file->raid_root;
  
  DBUG_ENTER("my_u_raid_close");
  DBUG_PRINT("enter",("MyFlags: %d", MyFlags)); 

  save_err= 0;
  my_chunks= my_raid_root->raid_vector;
  for (i= 0 ; i < my_raid_root->raid_chunks ; i++ )
  {
    err= my_simple_close((my_chunks + i)->my_file, MyFlags);
    if (err != 0)
    {
      save_err= err;
//      while (++i < my_raid_root->raid_chunks)
      free_my_file((my_chunks + i)->my_file);
//      my_free((char*)my_raid_root->raid_vector, MYF(0));
//      DBUG_RETURN(err);
    }
  }
//  my_free((char*)my_raid_root->raid_vector, MYF(0));
  DBUG_RETURN(save_err);
}
 
bool my_u_raid_create(MY_FILE * my_file, int CreateFlags,
		      int access_flags, myf MyFlags)
{
    /*
  DBUG_ENTER("my_u_raid_create");
  DBUG_PRINT("enter",
        ("Filename: %s  CreateFlags: %d  access_flags: %d  MyFlags: %d",
        my_file->name.buffer, CreateFlags, access_flags, MyFlags));
  if (raid_type)
  {
    DBUG_RETURN(intern_my_raid_create(my_file, CreateFlags, access_flags,
		raid_type, raid_chunks, raid_chunksize, MyFlags));
  }
  else
  {
    DBUG_RETURN(my_simple_create(my_file, CreateFlags, access_flags, MyFlags));
  }
  */
  return 1;
}

bool my_u_raid_open(MY_FILE * my_file, int Flags, myf MyFlags)
{
/*
  DBUG_ENTER("my_u_raid_open");
  DBUG_PRINT("enter",("Filename: %s  Flags: %d  MyFlags: %d",
                my_file->name.buffer, Flags, MyFlags));
  if (raid_type)
  {
    DBUG_RETURN(intern_my_raid_open(my_file, Flags, raid_type, raid_chunks, 
                raid_chunksize, MyFlags));
  }
  else
    DBUG_RETURN(my_simple_open (my_file, Flags, MyFlags));
*/
    return 1;    
}

size_t my_u_raid_pwrite(MY_FILE * my_file, const byte *Buffer, uint Count,
		      my_off_t offset, myf MyFlags)
{
  DBUG_ENTER("my_u_raid_pwrite");
  DBUG_PRINT("enter",("Fd: %d  Buffer: %lx  Count: %u offset: %u  MyFlags: %d",
		    my_file->file_descr, Buffer, Count, offset, MyFlags));

  /* Returning value isn't important because real seek is done later. */
  my_u_raid_seek(my_file, offset, MY_SEEK_SET, MyFlags);
  DBUG_RETURN(my_u_raid_write(my_file, Buffer, Count, MyFlags));
}


size_t my_u_raid_pread(MY_FILE * my_file, byte *Buffer, uint Count,
                   my_off_t offset, myf MyFlags)
{
  DBUG_ENTER("my_u_raid_pread");
  DBUG_PRINT("enter",("Fd: %d  Buffer: %lx  Count: %u offset: %u  MyFlags: %d",
		      my_file->file_descr, Buffer, Count, offset, MyFlags));

  /* Returning value isn't important because real seek is done later. */
  my_u_raid_seek(my_file, offset, MY_SEEK_SET, MyFlags);
  DBUG_RETURN(my_u_raid_read(my_file, Buffer, Count, MyFlags));
}

bool my_u_raid_rename(MY_FILE * from, MY_FILE * to,
		     uint raid_chunks, myf MyFlags)
{
  uint i;
  char from_tmp[FN_REFLEN];
  char to_tmp[FN_REFLEN];
  
  DBUG_ENTER("my_u_raid_rename");

  strmake(from_tmp, from->name.buffer, MF_DIR_LEN(from));
  strmake(to_tmp, to->name.buffer, MF_DIR_LEN(to));  
  for (i = 0 ; i < raid_chunks ; i++)
  {
    sprintf(from_tmp + MF_DIR_LEN(from), "%02x/%s", i, from->name.base);
    sprintf(to_tmp + MF_DIR_LEN(to), "%02x/%s", i, to->name.base);
    /* Convert if not unix */
    unpack_filename(from_tmp, from_tmp);
    unpack_filename(to_tmp,to_tmp);
    if (my_rename(from_tmp, to_tmp, MyFlags))
      DBUG_RETURN(1);
  }
  DBUG_RETURN(0);
}

bool my_u_raid_delete(MY_FILE * from, uint raid_chunks, myf MyFlags)
{
  char from_tmp[FN_REFLEN];
  uint i;
  DBUG_ENTER("my_u_raid_delete");

  for (i = 0 ; i < raid_chunks ; i++ )
  {
    strmake(from_tmp, from->name.buffer, MF_DIR_LEN(from));
    sprintf(from_tmp+MF_DIR_LEN(from), "%02x/%s", i, from->name.base);
    /* Convert if not unix */
//    unpack_filename(from_tmp, from_tmp);
    if (my_simple_delete(from, MyFlags))
	  DBUG_RETURN(1);
  }
  DBUG_RETURN(0);
}

bool my_u_raid_redel(MY_FILE * old_name, MY_FILE * new_name,
               uint raid_chunks, myf MyFlags)
{
  char new_name_buff[FN_REFLEN], old_name_buff[FN_REFLEN];
  char *new_end, *old_end;
  uint i,old_length,new_length;
  int error=0;
  DBUG_ENTER("my_u_raid_redel");

  old_end=old_name_buff+dirname_part(old_name_buff,old_name->name.buffer);
  old_length=MF_DIR_LEN(old_name);
  new_end=new_name_buff+dirname_part(new_name_buff,new_name->name.buffer);
  new_length=MF_DIR_LEN(new_name);
  for (i=0 ; i < raid_chunks ; i++)
  {
    MY_STAT status;
    sprintf(new_end, "%02x", i);
    if (my_stat(new_name_buff, &status, MYF(0)))
    {
      DBUG_PRINT("info",("%02x exists, skipping directory creation",i))
    }
    else
    {
	  if (my_mkdir(new_name_buff, 0777, MYF(0)))
	  {
	    DBUG_PRINT("error",("mkdir failed for %02x",i));
	    DBUG_RETURN(-1);
	  }
    }
    strxmov(strend(new_end),"/",new_name->name.buffer + new_length,NullS);
    sprintf(old_end, "%02x/%s", i, old_name->name.buffer + old_length);
    if (my_redel(old_name_buff, new_name_buff, MyFlags))
	error=1;
  }
  DBUG_RETURN(error);
}
