/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <my_global.h>
#include <my_dir.h>
#include <my_alarm.h>
#include <my_sys.h>
#include "m_string.h"
#include "mysys_err.h"
#include <errno.h>

my_bool init_tmp_file_template(MY_FILENAME *fn, 
			       const char *dir, const char *prefix)
{
  int dir_len= strlength(dir);
  int prefix_len= strlength(prefix);
  char *str_dst;
  const char *str_src;
  
  DBUG_ENTER("init_tmp_file_template");  
  if (!dir_len)
  {
    if (!prefix_len)
    {
      fn->base= fn->buffer= fn->ext= fn->dir= 0;
      fn->alloced_length= fn->buf_length= 0;
      fn->flags= 0;
    }
    else
    {
      fn->base= fn->buffer= my_memdup(prefix,prefix_len, MYF(0));
      fn->ext= fn->dir= 0;
      fn->alloced_length= fn->buf_length= prefix_len+1;
      fn->flags= MY_USE_ALLOCATED_BUFFER;
    }
  }
  else if (!prefix_len)
  {
    fn->buffer= my_memdup(dir,dir_len, MYF(0));
    fn->ext= fn->base= fn->dir= 0;
    fn->alloced_length= fn->buf_length= dir_len+1;
    fn->flags= MY_USE_ALLOCATED_BUFFER;
  }
  else
  {
    if (!(str_dst= fn->buffer= my_malloc(dir_len + prefix_len + 2, MYF(0))))
      DBUG_RETURN(1);
    fn->flags= MY_USE_ALLOCATED_BUFFER;
    
    fn->dir= str_dst;
    str_src= dir;
    do {
#ifdef FN_DEVCHAR
      if (*str_src==FN_DEVCHAR)
	fn->dir= str_dst+1;
#endif
      *str_dst++= *str_src++;
    } while(*str_src);
    
    if (*(str_dst-1)!=FN_LIBCHAR)
      *str_dst++= FN_LIBCHAR;
    
    fn->base= str_dst;
    
    str_src= prefix;
    do {
      *str_dst++= *str_src++;
    } while (*str_src);

    fn->ext= 0;
  }
  
  DBUG_RETURN(0);
}


MY_FILENAME *init_my_filename(MY_FILENAME * my_fname,
			      const char *fname,
			      uint16 init_flags)
{
  int len= strlength(fname);
  DBUG_ENTER("init_my_filename");
  
  if (!(init_flags & MY_USE_ALLOCATED_BUFFER))
    my_fname->buffer= (char*) fname;
  else
  {
    if (!(my_fname->buffer= my_memdup(fname, len+1, init_flags & MY_WME)))
      DBUG_RETURN(0);
    /* Remember that we have to free the buffer */
    my_fname->alloced_length= len;
  }
  
  my_fname->flags= init_flags;
  my_fname->buf_length= len;  
  
#ifdef FN_DEVCHAR
  if ((my_fname->dir=(char*)strrchr(my_fname->buffer,FN_DEVCHAR)) != 0)
    my_fname->dir++;
  else
#endif
    my_fname->dir= my_fname->buffer;
  
  if ((my_fname->base= strrchr(my_fname->dir, FN_LIBCHAR)))
    my_fname->base++;
  else
    my_fname->base= my_fname->dir;
  to_unix_path(my_fname->dir);
  
  if (!(my_fname->ext= strrchr(my_fname->base, FN_EXTCHAR)))
    my_fname->ext = my_fname->buffer + my_fname->buf_length;
  
  DBUG_RETURN(my_fname);
}

void free_my_filename(MY_FILENAME * fname)
{
  DBUG_ENTER("free_my_filename");
  if (fname->flags & MY_USE_ALLOCATED_BUFFER)
    my_free((char*) fname->buffer, MYF(0));
  DBUG_VOID_RETURN;
}

MY_FILE *init_my_file(MY_FILE *my_file, const char *fname, uint16 init_flags)
{
  DBUG_ENTER("init_my_file");
  if (init_flags & USE_ALLOCATED_MY_FILE || !my_file)
  {
    if (!(my_file= (MY_FILE* )my_malloc(sizeof(MY_FILE), MYF(0))))
      DBUG_RETURN(0);                            /* Not enough memory */
    init_flags|= USE_ALLOCATED_MY_FILE;
  }
  bzero((char*) my_file, sizeof(MY_FILE));
  
  if (!init_my_filename(&my_file->name, fname, init_flags))
  {
    free_my_file(my_file);
    DBUG_RETURN(0);
  }
  DBUG_PRINT("info",("ptr=%p filename=\"%s\"",my_file,my_file->name.buffer));

  DBUG_RETURN(my_file);
}

void free_my_file(MY_FILE * my_file)
{
  DBUG_ENTER("free_my_file");
  if (my_file->name.flags & MY_USE_ALLOCATED_BUFFER)
    my_free((char*) my_file->name.buffer, MYF(0));
  if (my_file->name.flags & USE_ALLOCATED_MY_FILE)
    my_free((char*) my_file, MYF(0));
  DBUG_VOID_RETURN;
}

void reinit_my_filename(MY_FILENAME * fname, char * buff, uint length)
{
  DBUG_ENTER("reinit_my_filename");

  if (fname->flags & MY_USE_ALLOCATED_BUFFER)
  {
    if (length > fname->alloced_length)
    {
      free_my_filename(fname);
      init_my_filename(fname, buff, MY_USE_ALLOCATED_BUFFER);
    }
    else
    {
      init_my_filename(fname, buff, 0);
    }
  }
  else
    init_my_filename(fname, buff, MY_USE_ALLOCATED_BUFFER);

  DBUG_VOID_RETURN;
}

bool replace_my_file_parts(MY_FILE *from, const char *dir,
			   const char *extension, uint flag)
{
  reg1 uint length, dev_length, dir_length, ext_length, path_length;
  char dev[FN_REFLEN], *pos, *full_name= 0;
  const char *ext;
  
  DBUG_ENTER("replace_my_file_parts");
  
  if ((length= MF_DIR_LEN(from)) == 0 || (flag & MY_REPLACE_DIR))
  {
    /* Use given directory */
    convert_dirname(dev,dir,NullS);             /* Fix to this OS */
  }
  else if ((flag & MY_RELATIVE_PATH) && !test_if_hard_path(from->name.buffer))
  {
    /* Put 'dir' before the given path */
    pos=convert_dirname(dev,dir,NullS);
    strmake(pos, from->name.buffer, length);
  }
  else
    strmake(dev, from->name.buffer, (length= MF_DIR_LEN(from)));
  
  if (flag & MY_PACK_FILENAME)
    pack_dirname(dev,dev);                      /* Put in ./.. and ~/.. */
  if (flag & MY_UNPACK_FILENAME)
    (void) unpack_dirname(dev,dev);             /* Replace ~/.. with dir */
  
  if (MF_EXT_LEN(from))
  {
    if ((flag & MY_REPLACE_EXT) == 0)           /* If we should keep old ext */
      ext= from->name.ext;                      /* Use old extension       */
    else
      ext= extension;                           /* Change extension        */
  }
  else
  {
    ext= extension;                             /* No ext, use the now one */
  }

  length= MF_NAME_LEN(from);
  ext_length= strlen(ext);
#ifdef FN_DEVCHAR
  if ((pos= (char*)strrchr(dev,FN_DEVCHAR)) != 0)
  {
    dev_length= pos - dev;
    dir_length= strlen(dev);
  }
  else
#endif
  {
    dev_length= 0;
    dir_length= strlen(dev);
  }
  path_length= dir_length + length + ext_length;
  
  if (path_length >= FN_REFLEN)				/* Very long path */
  {
    if (!(full_name= (char *) my_malloc(path_length + 1, MYF(0))))
        DBUG_RETURN(1);
    bmove(full_name, dev, dir_length);
    pos= strmake(full_name + dir_length, from->name.base, length);
#ifdef FN_UPPER_CASE
    caseup_str(full_name);
#endif
#ifdef FN_LOWER_CASE
    casedn_str(full_name);
#endif
    (void) strmov(pos,ext);
  }
  else							/* path < FN_REFLEN */
  {
    pos= strmake( dev + dir_length, from->name.base, length);
#ifdef FN_UPPER_CASE
    caseup_str(dev);
#endif
#ifdef FN_LOWER_CASE
    casedn_str(dev);
#endif
    (void) strmov(pos,ext);    
  }

  if ((from->name.flags & MY_USE_ALLOCATED_BUFFER) &&	/* We don't need  */
      (path_length <= from->name.alloced_length))       /* more memory    */
  {
    bmove(from->name.buffer, dev, path_length);
    from->name.buffer[path_length]= 0;
    from->name.buf_length= path_length;
  }
  else
  {
    if ((from->name.flags & MY_USE_ALLOCATED_BUFFER) && /* We need more memory*/
        (path_length > from->name.alloced_length))
      my_free((char*) from->name.buffer, MYF(0));

    if (full_name)					/* If path>=FN_REFLEN */
      from->name.buffer= full_name;
    else
      if (!(from->name.buffer= my_memdup(dev, path_length + 1, MYF(MY_WME))))
        DBUG_RETURN(1);
    from->name.buf_length= from->name.alloced_length= path_length;
    from->name.flags|= MY_USE_ALLOCATED_BUFFER;
  }

  from->name.dir= from->name.buffer + dev_length;
  from->name.base= from->name.dir + dir_length;
  from->name.ext= from->name.base + length;

  DBUG_RETURN(0);
}


int my_file_readlink(MY_FILENAME * fname, myf MyFlags)
{
#ifndef HAVE_READLINK
  return 1;
#else
  int result=0;
  int length;
  char buff[FN_REFLEN];
  DBUG_ENTER("my_filename_readlink");

  if ((length=readlink(fname->buffer, buff, FN_REFLEN-1)) < 0)
  {
    /* Don't give an error if this wasn't a symlink */
    if ((my_errno=errno) == EINVAL)
    {
      result= 1;
    }
    else
    {
      if (MyFlags & MY_WME)
        my_error(EE_CANT_READLINK, MYF(0), fname->buffer, errno);
      result= -1;
    }
  }
  else
  {
    buff[length]=0;
    reinit_my_filename(fname, buff, length);
  }
  DBUG_RETURN(result);
#endif /* HAVE_READLINK */
}


#if defined(SCO)
#define BUFF_LEN 4097
#elif defined(MAXPATHLEN)
#define BUFF_LEN MAXPATHLEN
#else
#define BUFF_LEN FN_LEN
#endif

int my_file_realpath(MY_FILENAME * fname, myf MyFlags)
{
#if defined(HAVE_REALPATH) && !defined(HAVE_purify) && !defined(HAVE_BROKEN_REALPATH)
  reg1 uint buf_length;
  int result=0;
  char buff[BUFF_LEN];
  struct stat stat_buff;
  DBUG_ENTER("my_file_realpath");
	
  if (!(MyFlags & MY_RESOLVE_LINK) ||
      (!lstat(fname->buffer, &stat_buff) && S_ISLNK(stat_buff.st_mode)))
  {
    DBUG_PRINT("info",("executing realpath"));
    if (realpath(fname->buffer, buff))
    {
      buf_length= strlen(buff);
      reinit_my_filename(fname, buff, buf_length);
    }
    else
    {
      /* Realpath didn't work;  Use original name */
      DBUG_PRINT("error",("realpath failed with errno: %d", errno));
      my_errno=errno;
      if (MyFlags & MY_WME)
          my_error(EE_REALPATH, MYF(0), fname->buffer, my_errno);
      result= -1;
    }
  }
  DBUG_RETURN(result);
#else
  return 0;
#endif
}

bool convert_my_filename(MY_FILE *from, uint flag)
{
  DBUG_ENTER("convert_my_filename");
  
  /*
    If MY_RETURN_REAL_PATH and MY_RESOLVE_SYMLINK is given, only do
    realpath if the file is a symbolic link
  */

  if (flag & MY_RETURN_REAL_PATH)
    (void) my_file_realpath(&from->name,
			    MYF(flag & MY_RESOLVE_SYMLINKS ? MY_RESOLVE_LINK: 0));
  else if (flag & MY_RESOLVE_SYMLINKS)
    (void) my_file_readlink(&from->name, MYF(0));
  else
    DBUG_RETURN(1);

  DBUG_RETURN(0);
}


char *system_my_filename(my_string to __attribute__((unused)), 
			 MY_FILE * my_file)
{
  DBUG_ENTER("system_my_filename");

#if FN_LIBCHAR != '/'				/* OS is not Unix */
  unpack_filename(to, my_file->name.buffer);
  DBUG_RETURN(to)
#else
  DBUG_RETURN(my_file->name.buffer);
#endif
}
