/* Copyright (C) 2002 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#define MY_USE_ALLOCATED_BUFFER  2048 /* Use allocated buffer for MY_FILENAME*/
#define USE_ALLOCATED_MY_FILE    4096 /* Use allocated MY_FILE struct        */
#define my_size_t uint

#ifdef __cplusplus
extern "C" {
#endif

/*
  structure for store parsed filename

  for example:

  d:/tmp/aa/exemple.txt
  ^ ^       ^       ^
  | |       |       |
  | |       |       |
  | dir     |       |
  | base----+       |
  | ext-------------+
  +-buffer
*/
typedef struct st_my_filename 
{
  char	  *dir;		      /* begin of directory name in buffer           */
  char	  *base;	      /* begin of file name in buffer                */
  char	  *ext;		      /* begin of file extention in buffer           */
  uint16  buf_length;	      /* length of directory name (without dev name) */
  uint16  alloced_length;     /* allocated buffer length                     */
  char    *buffer;            /* buffer, included full pathname..            */
  uint16  flags;              /* If we should free this on my_close and other*/    
} MY_FILENAME;

#define MF_DIR_LEN(info)  ((info)->name.base - (info)->name.buffer)
#define MF_NAME_LEN(info) ((info)->name.ext -  (info)->name.base)
#define MF_EXT_LEN(info)  ((info)->name.buffer + \
			   (info)->name.buf_length - (info)->name.ext)

enum en_file_type 
{
  EN_UNOPEN = 0,
  EN_FILE_BY_OPEN,
  EN_FILE_BY_CREATE,
  EN_STREAM_BY_FOPEN, 
  EN_STREAM_BY_FDOPEN, 
  EN_FILE_BY_MKSTEMP,
  EN_FILE_BY_DUP 
};


enum en_cache_type
{
  EN_UNDEF_CACHE = 0,
  EN_READ_CACHE,
  EN_WRITE_CACHE,
  EN_SEQ_READ_APPEND /* sequential read or append */,
  EN_READ_FIFO,
  EN_READ_NET,
  EN_WRITE_NET
};


enum en_raid_type 
{
  NONE, 
  STRIPPED
};


struct  st_my_file;
typedef int (*EN_IO_CACHE_CALLBACK)(struct st_my_file*);

#ifdef THREAD
typedef struct st_u_cache_share
{
  /* to sync on reads into buffer */
  pthread_mutex_t mutex;
  pthread_cond_t  cond;
  int             count;
  /* actual IO_CACHE that filled the buffer */
  struct st_my_file *active;
#ifdef NOT_YET_IMPLEMENTED
  /* whether the structure should be free'd */
  my_bool alloced;
#endif
} MY_CACHE_SHARE;
#endif

typedef struct st_my_file_cache
{

  /* 
     read_pos points to current read position in the buffer
     read_end is the non-inclusive boundary in the buffer for the currently
	      valid read area
     buffer is the read buffer
     not sure about request_pos except that it is used in async_io
  */
  /* Points to current read position in the buffer */
  byte *read_pos;
  /* the non-inclusive boundary in the buffer for the currently valid read */
  byte *read_end;
  byte *buffer;             /* The read buffer */
  byte *request_pos;        /* Used in ASYNC_IO */

  /* Only used in WRITE caches and in SEQ_READ_APPEND to buffer writes */
  byte *write_buffer;
  /*
    Only used in SEQ_READ_APPEND, and points to the current read position
    in the write buffer. Note that reads in SEQ_READ_APPEND caches can
    happen from both read buffer (byte* buffer) and write buffer
    (byte* write_buffer).
  */
  byte *append_read_pos; 
  /* Points to current write position in the write buffer */
  byte *write_pos;
  /* The non-inclusive boundary of the valid write area */
  byte *write_end;

  /* 
     current_pos and current_end are convenience variables used by
     my_b_tell() and other routines that need to know the current offset
     current_pos points to &write_pos, and current_end to &write_end in a
     WRITE_CACHE, and &read_pos and &read_end respectively otherwise
  */
  byte **current_pos;
  byte **current_end;

/* The lock is for append buffer used in SEQ_READ_APPEND cache */
#ifdef THREAD
  /*
    The lock is for append buffer used in SEQ_READ_APPEND cache
    need mutex copying from append buffer to read buffer.
  */
  pthread_mutex_t append_buffer_lock;
  /*
    The following is used when several threads are reading the
    same file in parallel. They are synchronized on disk
    accesses reading the cached part of the file asynchronously.
    It should be set to NULL to disable the feature.  Only
    READ_CACHE mode is supported.
  */
  struct st_u_cache_share *share;
#endif

  /* 
     specifies the type of the cache. Depending on the type of the cache
     certain operations might not be available and yield unpredicatable
     results. Details to be documented later
  */
  enum en_cache_type type;

  /* 
     callbacks when the actual read I/O happens. These were added and
     are currently used for binary logging of LOAD DATA INFILE - when a
     block is read from the file, we create a block create/append event, and
     when IO_CACHE is closed, we create an end event. These functions could,
     of course be used for other things
  */
  EN_IO_CACHE_CALLBACK pre_read;
  EN_IO_CACHE_CALLBACK post_read;
  EN_IO_CACHE_CALLBACK pre_close;
  void* arg; /* for use by pre/post_read */

  /* 
     seek_not_done is set by my_b_seek() to inform the upcoming read/write
     operation that a seek needs to be preformed prior to 
     the actual I/O
  */
  int seek_not_done;
  /*
    error is 0 if the cache operation was successful, -1 if there was a
    "hard" error, and the actual number of I/O-ed bytes if the 
    read/write was partial
  */
  int error;

  /* buffer_length is memory size allocated for buffer or write_buffer */
  uint buffer_length;
  /* read_length is the same as buffer_length except when we use async io */
  uint read_length;

  /* 
     alloced_buffer is 1 if the buffer was allocated by init_io_cache() and
     0 if it was supplied by the user
     Currently READ_NET is the only one that will use a buffer allocated
     somewhere else
  */
  my_bool alloced_buffer;

#ifdef HAVE_AIOWAIT
  /*
    As inidicated by ifdef, this is for async I/O, which is not currently
    used (because it's not reliable on all systems)
  */
  uint inited;
  my_off_t aio_read_pos;
  my_aio_result aio_result;
#endif

} MY_FILE_CACHE;


typedef struct st_my_file_raid_root
{
  ulong                 raid_chunksize;             /* 1..n in bytes */
  enum en_raid_type     raid_type;
  uint                  raid_chunks;
  struct my_chunk       *raid_vector;   
} MY_FILE_RAID_ROOT;


typedef struct st_my_file
{
  File		      file_descr; /* file descriptor			     */
  enum en_file_type   file_type;  /* type of file			     */
  MY_FILE_CACHE	      cache;	  /* cache				     */
  MY_FILE_RAID_ROOT   raid_root;  /* root struct for raid file		     */
  myf		      myflags;	  /* Flags used to my_read/my_write	     */
  struct my_file_vmt  *vmt;	  /* table of virtual methods		     */

  /* Offset in file corresponding to the first byte of byte* buffer. */
  my_off_t pos_in_file;
  /*
    The offset of end of file for READ_CACHE and WRITE_CACHE.
    For SEQ_READ_APPEND it the maximum of the actual end of file and
    the position represented by read_end.
  */
  my_off_t end_of_file;
  
#if defined(THREAD) && !defined(HAVE_PREAD)  
  pthread_mutex_t	mutex;
#endif

  MY_FILENAME	      name;	  /* file name */
} MY_FILE;


typedef struct my_chunk
{
  MY_FILE* my_file;
  my_off_t pos_in_chunk;
} MY_CHUNK;


struct my_stat;
typedef struct my_file_vmt
{
  struct my_file_vmt *parent_vmt;

  my_size_t   (*read)   (MY_FILE *,       byte *, my_size_t,	       myf);
  my_size_t   (*pread)  (MY_FILE *,       byte *, my_size_t, my_off_t, myf);
  my_size_t   (*write)  (MY_FILE *, const byte *, my_size_t,	       myf);
  my_size_t   (*pwrite) (MY_FILE *, const byte *, my_size_t, my_off_t, myf);

  int      (*close)  (MY_FILE *,				 myf);
  int      (*lock)   (MY_FILE *, int locktype, 
			       my_off_t start, my_off_t length,  myf);
  int      (*chsize) (MY_FILE *, my_off_t newlength, int filler, myf);
  int      (*fstat)  (MY_FILE *, struct my_stat*,		 myf);

  my_off_t (*mtell)  (MY_FILE *,				 myf);
  my_off_t (*seek)   (MY_FILE *, my_off_t pos, int whence,	 myf);

  void     (*free)   (MY_FILE *);
} MY_FILE_VMT;

/* Standard methods for work with MY_FILE: */
#define my_u_read_bytes(f,b,c,fl)\
((f)->cache.read_pos + (c) <= (f)->cache.read_end\
 ?(memcpy(b,(f)->cache.read_pos,(my_size_t) (c)), ((f)->cache.read_pos+=(c)),(c))\
 :((f)->vmt->read((f),(b),(c),(fl))))
#define my_u_pread_bytes(f,b,c,p,fl)((f)->vmt->pread((f),(b),(c),(p),(fl)))
#define my_u_write_bytes(f,b,c,fl)\
((f)->cache.write_pos + (c) <= (f)->cache.write_end \
 ?(memcpy((f)->cache.write_pos,b,(my_size_t)(c)),((f)->cache.write_pos+=(c)),(c))\
 :((f)->vmt->write((f),(b),(c),(fl))))
#define my_u_pwrite_bytes(f,b,c,p,d)((f)->vmt->pwrite((f),(b),(c),(p),(d)))

#define my_u_read(f,b,c,fl)\
((f)->cache.read_pos + (c) <= (f)->cache.read_end\
 ?(memcpy(b,(f)->cache.read_pos,(my_size_t) (c)), ((f)->cache.read_pos+=(c)),0)\
 :(((f)->vmt->read((f),(b),(c),(fl)))!=(c)))
#define my_u_pread(f,b,c,p,fl)\
 (((f)->vmt->pread((f),(b),(c),(p),(fl)))!=(c))
#define my_u_write(f,b,c,fl)\
((f)->cache.write_pos + (c) <= (f)->cache.write_end \
 ?(memcpy((f)->cache.write_pos,b,(my_size_t)(c)),((f)->cache.write_pos+=(c)),0)\
 :(((f)->vmt->write((f),(b),(c),(fl)))!=(c)))
#define my_u_pwrite(f,b,c,p,d)\
  (((f)->vmt->pwrite((f),(b),(c),(p),(d)))!=(c))

#define my_u_close(f,fl)((!(f)||!(f)->vmt)?0:(f)->vmt->close(f,fl))
#define my_u_lock(f,lt,s,l,fl)((f)->vmt->lock((f),(lt),(s),(l),(fl)))
#define my_u_chsize(f,nl,l,fl)((f)->vmt->chsize((f),(nl),(l),(fl)))
#define my_u_fstat(f,sa,fl)((f)->vmt->fstat((f),(sa),(fl)))
#define my_u_tell(f,fl)((f)->vmt->mtell((f),(fl)))
#define my_u_seek(f,p,w,fl)((f)->vmt->seek((f),(p),(w),(fl)))

#define my_u_bytes_in_cache(f)\
  (uint) (*(f)->cache.current_end - \
	  *(f)->cache.current_pos)

#define my_u_get(info) \
  ((info)->cache.read_pos != (info)->cache.read_end ?\
   ((info)->cache.read_pos++, (int) (uchar) (info)->cache.read_pos[-1]) :\
   _my_u_get(info))

/* Utilites for work with MY_FILE */
extern MY_FILE *init_my_file(MY_FILE *my_file, const char *fname, uint16 init_flags);
extern void     free_my_file(MY_FILE *my_file);

extern int _my_u_get(MY_FILE*);  /* read once byte from file */

/* my_file extra macro */

#define my_u_super_read(f,b,c,fl)(f->vmt->parent_vmt->read((f),(b),(c),(fl)))
#define my_u_super_pread(f,b,c,p,fl)\
			    (f->vmt->parent_vmt->pread((f),(b),(c),(p),(fl)))
#define my_u_super_write(f,b,c,fl)(f->vmt->parent_vmt->write((f),(b),(c),(fl)))
#define my_u_super_pwrite(f,b,c,p,fl)(f->vmt->parent_vmt->pwrite((f),(b),(c),(p),(fl)))
#define my_u_super_close(f,fl)(f->vmt->parent_vmt->close(f,fl))
#define my_u_super_lock(f,lt,s,l,fl)\
			    (f->vmt->parent_vmt->lock((f),(lt),(s),(l),(fl)))
#define my_u_super_chsize(f,nl,l,fl)(f->vmt->parent_vmt->chsize((f),(nl),(l),(fl)))
#define my_u_super_fstat(f,sa,fl)(f->vmt->parent_vmt->fstat((f),(sa),(fl)))
#define my_u_super_tell(f,fl)(f->vmt->parent_vmt->mtell((f),(fl)))
#define my_u_super_seek(f,p,w,fl)(f->vmt->parent_vmt->seek((f),(p),(w),(fl)))

my_bool init_tmp_file_template(MY_FILENAME *fn, 
			       const char *dir, const char *prefix);

/* simple file special functions */

extern bool my_simple_create(MY_FILE * my_file, int CreateFlags,
			     int access_flags, myf MyFlags);
extern bool my_simple_open(MY_FILE * my_file, int Flags, myf MyFlags);
extern int  my_simple_delete(MY_FILE * fs, myf MyFlags);

/* raid file special functions */

extern bool my_u_raid_open   (MY_FILE * my_file, int Flags, myf MyFlags);
extern bool my_u_raid_create (MY_FILE * my_file, int CreateFlags, int access_flags,
			      myf MyFlags);
extern bool my_u_raid_rename (MY_FILE * from, MY_FILE * to, uint raid_chunks, myf MyFlags);
extern bool my_u_raid_delete (MY_FILE * from, uint raid_chunks, myf MyFlags);
extern bool my_u_raid_redel  (MY_FILE * old_name, MY_FILE * new_name, uint raid_chunks,
			      myf MyFlags);

/* cache file special functions */

extern int init_u_cache(MY_FILE *fs, uint cachesize,
			enum en_cache_type type, my_off_t seek_offset,
			pbool use_async_io, myf cache_myflags);
extern my_bool reinit_u_cache(MY_FILE *fs, enum en_cache_type type,
			      my_off_t seek_offset,
			      pbool use_async_io __attribute__((unused)),
			      pbool clear_cache);
extern int end_u_cache(MY_FILE *file);
extern int my_u_cache_append(MY_FILE *file, const byte *Buffer, uint Count);
extern int my_u_safe_write(MY_FILE *file, const byte *Buffer, uint Count);

#define flush_u_cache(fs) _flush_u_cache((fs),1)
extern int _flush_u_cache(MY_FILE *fs, int need_append_buffer_lock);

extern int my_u_block_write(register MY_FILE *info, const byte *Buffer, 
			    my_size_t Count, my_off_t pos, myf myflags);
extern MY_FILE *u_open_cached_file(MY_FILE *file, const char* dir, 
				   const char *prefix, uint cache_size, 
				   myf cache_myflags);

#ifdef THREAD
extern void init_u_cache_share(MY_FILE *file, 
			       MY_CACHE_SHARE *s, uint num_threads);
extern void remove_u_thread(MY_FILE *file);
#endif

/* simple file common functions */

extern my_size_t my_simple_read  (MY_FILE*,	 byte*,my_size_t,	  myf);
extern my_size_t my_simple_pread (MY_FILE*,	 byte*,my_size_t,my_off_t,myf);
extern my_size_t my_simple_write (MY_FILE*,const byte*,my_size_t,	  myf);
extern my_size_t my_simple_pwrite(MY_FILE*,const byte*,my_size_t,my_off_t,myf);
  
extern int my_simple_close  (MY_FILE *,				      myf);
extern int my_simple_lock   (MY_FILE *, 
			     int locktype, my_off_t start, 
			     my_off_t length,			      myf);
extern int my_simple_chsize (MY_FILE *, my_off_t newlength, int fill, myf);
extern int my_simple_fstat  (MY_FILE *, struct my_stat *,	      myf);
  
extern my_off_t my_simple_tell(MY_FILE *,			      myf);
extern my_off_t my_simple_seek(MY_FILE *, my_off_t pos, int whence,   myf);
  
extern uint my_simple_fill    (MY_FILE *);

extern void my_simple_free    (MY_FILE *);
  
extern MY_FILE_VMT simple_file_vmt;

/* raid file common functions */

extern size_t my_u_raid_read  (MY_FILE *,	byte *, size_t,		  myf);
extern size_t my_u_raid_pread (MY_FILE *,	byte *, size_t, my_off_t, myf);
extern size_t my_u_raid_write (MY_FILE *, const byte *, size_t,		  myf);
extern size_t my_u_raid_pwrite(MY_FILE *, const byte *, size_t, my_off_t, myf);

extern int my_u_raid_close  (MY_FILE *,				      myf);
extern int my_u_raid_lock   (MY_FILE *, int locktype, 
			    my_off_t start, my_off_t length,	      myf);
extern int my_u_raid_chsize (MY_FILE *, my_off_t newlength, int filler, myf);
extern int my_u_raid_fstat  (MY_FILE *, struct my_stat*,	      myf);
my_off_t my_u_raid_tell	    (MY_FILE *,				      myf);
my_off_t my_u_raid_seek	    (MY_FILE *, my_off_t pos, int whence,     myf);
extern uint my_u_raid_fill  (MY_FILE *);

extern void my_u_raid_free  (MY_FILE *);

extern MY_FILE_VMT raid_file_vmt;

/* cache file common functions */

extern my_size_t my_cache_read    (MY_FILE*,      byte*,my_size_t,	   myf);
#ifdef HAVE_AIOWAIT
extern int my_cache_read_async    (MY_FILE*,      byte*,my_size_t,	   myf);
#endif
extern my_size_t my_cache_read_net(MY_FILE*,      byte*,my_size_t,	   myf);
extern my_size_t my_cache_seq_read(MY_FILE*,      byte*,my_size_t,	   myf);
extern my_size_t my_cache_pread   (MY_FILE*,      byte*,my_size_t,my_off_t,myf);
extern my_size_t my_cache_write   (MY_FILE*,const byte*,my_size_t,	   myf);
extern my_size_t my_cache_pwrite  (MY_FILE*,const byte*,my_size_t,my_off_t,myf);

extern int my_cache_close      (MY_FILE *,				  myf);
extern int my_cache_chsize     (MY_FILE *, my_off_t newlength, int filler,myf);

extern my_off_t my_cache_tell  (MY_FILE *,				  myf);
extern my_off_t my_cache_seek  (MY_FILE *, my_off_t pos, int whence,	  myf);
extern uint my_cache_fill      (MY_FILE *);

extern void my_cache_free      (MY_FILE *);

extern my_off_t my_cache_append_tell(MY_FILE *file, myf);

extern MY_FILE_VMT cache_file_vmt;
extern MY_FILE_VMT cache_read_file_vmt;
extern MY_FILE_VMT cache_write_file_vmt;
extern MY_FILE_VMT cache_seq_read_append_file_vmt;

/* myfile service functions */

extern uint my_cache_fill(MY_FILE *info);
extern uint my_u_cache_filelength(MY_FILE *info);
extern uint my_u_gets(MY_FILE *info, char *to, uint max_length);
extern uint my_u_printf(MY_FILE *info, const char* fmt, ...);
extern uint my_u_vprintf(MY_FILE *info, const char* fmt, va_list ap);

#ifdef __cplusplus
}
#endif
