/* Copyright (C) 2002 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "my_global.h"
#include "my_sys.h"
#include "m_string.h"

uint my_u_cache_filelength(MY_FILE *info)
{
  if (info->cache.type == EN_WRITE_CACHE)
  {
    return my_u_tell(info,MYF(0));
  }
  else
  {
    info->cache.seek_not_done=1;
    return my_u_super_seek(info,0L,MY_SEEK_END,MYF(0));
  }
  return 0;
}

/*
  Read a string ended by '\n' into a buffer of 'max_length' size.
  Returns number of characters read, 0 on error.
  last byte is set to '\0'
  If buffer is full then to[max_length-1] will be set to \0.
*/
uint my_u_gets(MY_FILE *file, char *to, uint max_length)
{
  MY_FILE_CACHE *info = &file->cache;
  char *start = to;
  uint length;
  DBUG_ENTER("my_u_gets");
  
  max_length--;					/* Save place for end \0 */
  /* Calculate number of characters in buffer */
  if (!(length= my_u_bytes_in_cache(file)) && 
      !(length= my_cache_fill(file)))
    DBUG_RETURN(0);
  for (;;)
  {
    char *pos,*end;
    if (length > max_length)
      length=max_length;
    for (pos=info->read_pos,end=pos+length ; pos < end ;)
    {
      if ((*to++ = *pos++) == '\n')
      {
	info->read_pos=pos;
	*to='\0';
	DBUG_RETURN((uint) (to-start));
      }
    }
    if (!(max_length-=length))
    {
      /* Found enough charcters;  Return found string */
      info->read_pos=pos;
      *to='\0';
      DBUG_RETURN((uint) (to-start));
    }
    if (!(length=my_cache_fill(file)))
      DBUG_RETURN(0);
  }
}


/*
  Simple printf version.  Supports '%s', '%d', '%u', "%ld" and "%lu"
  Used for logging in MySQL
  returns number of written character, or (uint) -1 on error
*/
uint my_u_printf(MY_FILE *info, const char* fmt, ...)
{
  int result;
  va_list args;
  DBUG_ENTER("my_u_printf");
  va_start(args,fmt);
  result=my_u_vprintf(info, fmt, args);
  va_end(args);
  DBUG_RETURN(result);
}

uint my_u_vprintf(MY_FILE *file, const char* fmt, va_list args)
{
  uint out_length=0;
  DBUG_ENTER("my_u_vprintf");
  
  for (; *fmt ; fmt++)
  {
    if (*fmt++ != '%')
    {
      /* Copy everything until '%' or end of string */
      const char *start=fmt-1;
      uint length;
      for (; *fmt && *fmt != '%' ; fmt++ ) ;
      length= (uint) (fmt - start);
      out_length+=length;
      if (my_u_write(file, start, length, MYF(0)))
	goto err;
      if (!*fmt)				/* End of format */
      {
	DBUG_RETURN(out_length);
      }
      fmt++;
      /* Found one '%' */
    }
    /* Skipp if max size is used (to be compatible with printf) */
    while (my_isdigit(&my_charset_latin1, *fmt) || *fmt == '.' || *fmt == '-')
      fmt++;
    if (*fmt == 's')				/* String parameter */
    {
      reg2 char *par = va_arg(args, char *);
      uint length = (uint) strlen(par);
      out_length+=length;
      if (my_u_write(file, par, length, MYF(0)))
	goto err;
    }
    else if (*fmt == 'd' || *fmt == 'u')	/* Integer parameter */
    {
      register int iarg;
      uint length;
      char buff[17];
      
      iarg = va_arg(args, int);
      if (*fmt == 'd')
	length= (uint) (int10_to_str((long) iarg,buff, -10)- buff);
      else
	length= (uint) (int10_to_str((long) (uint) iarg,buff,10)- buff);
      out_length+=length;
      if (my_u_write(file, buff, length, MYF(0)))
	goto err;
    }
    else if ((*fmt == 'l' && fmt[1] == 'd') || fmt[1] == 'u')
      /* long parameter */
    {
      register long iarg;
      uint length;
      char buff[17];
      
      iarg = va_arg(args, long);
      if (*++fmt == 'd')
	length= (uint) (int10_to_str(iarg,buff, -10) - buff);
      else
	length= (uint) (int10_to_str(iarg,buff,10)- buff);
      out_length+=length;
      if (my_u_write(file, buff, length, MYF(0)))
	goto err;
    }
    else
    {
      /* %% or unknown code */
      if (my_u_write(file, "%", 1, MYF(0)))
	goto err;
      out_length++;
    }
  }
  DBUG_RETURN(out_length);
  
err:
  DBUG_RETURN((uint) -1);
}

int _my_u_get(MY_FILE* fs)
{
  byte buff;
  EN_IO_CACHE_CALLBACK pre_read,post_read;
  DBUG_ENTER("_my_u_get");
  if ((pre_read = fs->cache.pre_read))
    (*pre_read)(fs);
  if ((*(fs)->vmt->read)(fs,&buff,1,MYF(0)))
    DBUG_RETURN(my_b_EOF);
  if ((post_read = fs->cache.post_read))
    (*post_read)(fs);
  DBUG_RETURN((int) (uchar) buff);
}

/*
  Fill buffer.  Note that this assumes that you have already used
  all characters in the CACHE, independent of the read_pos value!
  return:  0 on error or EOF (info->error = -1 on error)
  number of characters
*/
uint my_cache_fill(MY_FILE *file)
{
  MY_FILE_CACHE *info= &file->cache;
  my_off_t pos_in_file=(file->pos_in_file+
			(uint) (info->read_end - info->buffer));
  my_off_t max_length;
  uint diff_length,length;
  DBUG_ENTER("my_cache_fill");
  if (info->seek_not_done)
  {					/* File touched, do seek */
    if (my_u_super_seek(file,pos_in_file,MY_SEEK_SET,MYF(0)) ==
	MY_FILEPOS_ERROR)
    {
      info->error= 0;
      DBUG_RETURN(0);
    }
    info->seek_not_done=0;
  }
  diff_length=(uint) (pos_in_file & (IO_SIZE-1));
  max_length= (my_off_t) (file->end_of_file - pos_in_file);
  if (max_length > (my_off_t) (info->read_length-diff_length))
    max_length=(my_off_t) (info->read_length-diff_length);
  if (!max_length)
  {
    info->error= 0;
    DBUG_RETURN(0);					/* EOF */
  }
  else if ((length=my_u_super_read(file,info->buffer,max_length,
				   file->myflags))!=max_length)
  {
    info->error= -1;
    DBUG_RETURN(0);
  }
  info->read_pos=info->buffer;
  info->read_end=info->buffer+length;
  file->pos_in_file=pos_in_file;
  DBUG_RETURN(length);
}

uint my_cache_length(MY_FILE *info)
{
  if (info->cache.type == EN_WRITE_CACHE)
  {
    return my_u_tell(info,MYF(0));
  }
  else
  {
    info->cache.seek_not_done=1;
    return my_u_super_seek(info,0L,MY_SEEK_END,MYF(0));
  }
}
